# -*- coding: utf-8 -*-
"""
colibri.channel
~~~~~~~~~~~~~~~

AMQP channels.
"""

from functools import partial

from asyncio import coroutine

from . import spec
from .frame import FrameHandler
from .message import BasicMessage
from .utils import YieldFromContextManager
from .exceptions import QueueEmpty


class Channel(object):
    """This class represents an AMQP channel"""

    def __init__(self, connection, channel_id=None, on_open=None):
        if channel_id:
            connection._claim_channel_id(channel_id)
        else:
            channel_id = connection._get_free_channel_id()

        self.channel_id = channel_id
        self.connection = connection
        self.protocol = connection.protocol
        self.loop = connection.loop

        self.is_open = False
        self.active = True

        self.consumer_callbacks = {}
        self.cancel_callbacks = {}
        self.no_ack_consumers = set()
        self.events = set()

        self._confirm_selected = False

        self.handler = FrameHandler(self)
        self.handle = self.handler.handle

        # In case connection confirms publications
        if connection.confirm_publish:
            self.basic_publish = self.basic_publish_confirm

        self.on_open = on_open
        self.send_method = partial(connection.send_method,
                                   channel_id=channel_id)

        self._setup_callbacks()

    def _setup_callbacks(self):
        self.handler.update({
            spec.ChannelClose: self._on_close,
            spec.ChannelCloseOK: self._on_close_ok,
            spec.ChannelFlow: self._on_flow,
            spec.ChannelOpenOK: self._on_open_ok,
            spec.BasicCancel: self._on_basic_cancel,
            spec.BasicCancelOK: self._on_basic_cancel_ok,
            spec.BasicDeliver: self._on_basic_deliver,
            spec.BasicReturn: self._on_basic_return,
            spec.BasicAck: self._on_basic_ack,
        })

    def open(self):
        """Open channel"""
        return (yield from self._do_open())

    def close(self, reply_code=0, reply_text='', class_id=0, method_id=0,
              method=spec.ChannelClose, reply_method=spec.ChannelCloseOK):
        """Gracefully close channel"""
        if not self.is_open or self.connection is None:
            return
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (reply_code, reply_text, class_id, method_id),
            )
            res = yield from fut
            self._do_close()
            return res

    def __enter__(self):
        raise RuntimeError(
            '"yield from" should be used as context manager exception')

    def __exit__(self, *args):
        pass

    def __iter__(self):
        # This is not a coroutine.  It is meant to enable the idiom:
        #
        #     with (yield from chan):
        #         <block>
        #
        # as an alternative to:
        #
        #     yield from chan.open()
        #     try:
        #         <block>
        #     finally:
        #         chan.close()
        yield from self.open()
        return YieldFromContextManager(self)

    @coroutine
    def _on_close(self, reply_code, reply_text, class_id, method_id,
                  method=spec.ChannelCloseOK):
        self.send_method(method, ())
        yield from self.channel._do_revive()

    @coroutine
    def _on_close_ok(self):
        self._do_close()
        self.handler.ready()

    @coroutine
    def _on_flow(self, active, method=spec.ChannelFlowOK):
        self.active = active
        self.send_method(method, (active,))

    def _do_open(self, method=spec.ChannelOpen,
                 reply_method=spec.ChannelOpenOK):
        if self.is_open:
            return
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, ('',))
            return (yield from fut)

    @coroutine
    def _on_open_ok(self, reserved_1):
        self.is_open = True
        self.on_open and self.on_open(self)
        self.handler.ready()

    @coroutine
    def _on_basic_cancel(self, consumer_tag):
        callback = self._remove_tag(consumer_tag)
        if callback:
            callback(consumer_tag)
        else:
            print('on basic cancel')
            # raise ConsumerCancelled(consumer_tag, spec.Basic.Cancel)

    @coroutine
    def _on_basic_cancel_ok(self, consumer_tag):
        self._remove_tag(consumer_tag)
        self.handler.ready()

    def _remove_tag(self, consumer_tag):
        self.consumer_callbacks.pop(consumer_tag, None)
        self.cancel_callbacks.pop(consumer_tag, None)

    @coroutine
    def _on_basic_deliver(self, consumer_tag, delivery_tag, redelivered,
                          exchange, routing_key, msg):
        msg.channel = self
        msg.delivery_info = {
            'consumer_tag': consumer_tag,
            'delivery_tag': delivery_tag,
            'redelivered': redelivered,
            'exchange': exchange,
            'routing_key': routing_key,
        }
        try:
            fun = self.consumer_callbacks[consumer_tag]
        except KeyError:
            pass
        else:
            fun(msg)

    @coroutine
    def _on_basic_return(self, reply_code, reply_text,
                         exchange, routing_key, message):
        # exc = error_for_code(
        #     reply_code, reply_text, spec.Basic.Return, ChannelError,
        # )
        # TODO work with events
        exc = None
        handlers = self.events.get('basic_return')
        if not handlers:
            raise exc
        for callback in handlers:
            callback(exc, exchange, routing_key, message)

    @coroutine
    def _on_basic_ack(self, delivery_tag, multiple):
        for callback in self.events['basic_ack']:
            callback(delivery_tag, multiple)

    def _do_close(self):
        self.is_open = False
        channel_id, self.channel_id = self.channel_id, None
        connection, self.connection = self.connection, None
        if connection:
            connection.channels.pop(channel_id, None)
            connection._available_channel_ids.append(channel_id)

    def _do_revive(self):
        self.is_open = True
        return (yield from self._do_open())

    # Public API goes below

    def flow(self, active, *, method=spec.ChannelFlow,
             reply_method=spec.ChannelFlowOK):
        """Ask the peer to pause or restart the flow of content
        data sent by a consumer. This is a simple flow-control mechanism
        that a peer can use to avoid overflowing its queues or otherwise
        finding itself receiving more messages than it can process.
        Note that this method is not intended for window control. It does
        not affect contents returned by Basic.Get-Ok methods.

        :param active: Specify False to pause the flow and True to resume.

        """
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (active,))
            return (yield from fut)

    # Work with Exchanges

    def exchange_declare(self, exchange, type, *, passive=False, durable=False,
                         auto_delete=False, internal=False, no_wait=False,
                         arguments=None, method=spec.ExchangeDeclare,
                         reply_method=spec.ExchangeDeclareOK):
        """Create an exchange if it does not already exist,
        and if the exchange exists, verify that it is
        of the correct and expected class.

        :param exchange: Exchange name
        :param type: Exchange type - direct, topic, fanout or headers
        :keyword passive: If True, the server will reply with ExchangeDeclareOK
                          if the exchange already exists with the same name,
                          and raise an error if not.
        :keyword durable: If True when creating a new exchange, the exchange
                          will be marked as durable.  Durable exchanges
                          remain active when a server restarts.
        :keyword auto_delete: If True, the exchange is deleted when all queues
                              have finished using it.
        :keyword internal: If True, the exchange may not be used directly by
                           publishers, but only when bound to other exchanges.
        :keyword no_wait: If True, the client will not wait for the reply.
        :keyword arguments: Additional arguments.
        """
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, exchange, type, passive, durable, auto_delete,
                 internal, no_wait, arguments),
            )
            return (yield from fut)

    def exchange_delete(self, exchange, *, if_unused=False, no_wait=False,
                        method=spec.ExchangeDelete,
                        reply_method=spec.ExchangeDeleteOK):
        """Delete an exchange. When an exchange is deleted
        all queue bindings on the exchange are cancelled.

        :param exchange: Exchange name.
        :keyword if_unused: If True, the server will only delete the exchange
                            if it has no queue bindings.
        :keywork no_wait: If True, the client will not wait for the reply.
        """
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (0, exchange, if_unused, no_wait))
            return (yield from fut)

    def exchange_bind(self, destination, source, *, routing_key=None,
                      no_wait=False, arguments=None,
                      method=spec.ExchangeBind,
                      reply_method=spec.ExchangeBindOK):
        """Bind an exchange to an exchange.

        :param destination: The name of the destination exchange to bind.
        :param source: The name of the source exchange to bind.
        :keyword routing_key: The routing key for the binding.
        :keywork no_wait: If True, the client will not wait for the reply.
        :keyword arguments: Additional arguments.
        """
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, destination, source, routing_key,
                 no_wait, arguments),
            )
            return (yield from fut)

    def exchange_unbind(self, destination, source, *, routing_key=None,
                        no_wait=False, arguments=None,
                        method=spec.ExchangeUnbind,
                        reply_method=spec.ExchangeUnbindOK):
        """Unbind an exchange from an exchange.

        :param destination: The name of the destination exchange to unbind.
        :param source: The name of the source exchange to unbind.
        :keyword routing_key: The routing key to unbind.
        :keywork no_wait: If True, the client will not wait for the reply.
        :keyword arguments: Additional arguments.
        """
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, destination, source, routing_key, no_wait, arguments),
            )
            return (yield from fut)

    # Work with Queues

    def queue_declare(self, queue, *, passive=False, durable=False,
                      exclusive=False, auto_delete=True, no_wait=False,
                      arguments=None, method=spec.QueueDeclare,
                      reply_method=spec.QueueDeclareOK):
        """Create or check a queue.

        :param queue: Queue name.
        :keyword passive: If True, the server will reply with QueueDeclareOK
                          if the queue already exists with the same name, and
                          raise an error if not.
        :keyword durable: If True when creating a new queue, the queue
                          will be marked as durable. Durable queues
                          remain active when a server restarts.
        :keyword exclusive: If True, the queue will be marked as exclusive.
                            Exclusive queues may only be accessed
                            by the current connection, and are deleted
                            when that connection closes.
        :keyword auto_detele: If True, the queue is deleted when all consumers
                              have finished using it.
        :keywork no_wait: If True, the client will not wait for the reply.
        :keyword arguments: Additional arguments.
        """
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, queue, passive, durable, exclusive, auto_delete,
                 no_wait, arguments),
            )
            return (yield from fut)

    def queue_bind(self, queue, exchange, routing_key, *,
                   no_wait=False, arguments=None,
                   method=spec.QueueBind,
                   reply_method=spec.QueueBindOK):
        """Bind a queue to an exchange.

        :param queue: The name of the queue to bind.
        :param exchange: The exchange name to bind the queue to.
        :param routing_key: The routing key for the binding.
        :keywork no_wait: If True, the client will not wait for the reply.
        :keyword arguments: Additional arguments.
        """
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, queue, exchange, routing_key, no_wait, arguments),
            )
            return (yield from fut)

    def queue_unbind(self, queue, exchange, *, routing_key=None, no_wait=False,
                     arguments=None, method=spec.QueueUnbind,
                     reply_method=spec.QueueUnbindOK):
        """Unbind a queue from an exchange.

        :param queue: The name of the queue to unbind.
        :param exchange: The exchange name to unbind the queue from.
        :keyword routing_key: The routing key to unbind.
        :keywork no_wait: If True, the client will not wait for the reply.
        :keyword arguments: Additional arguments.
        """
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, queue, exchange, routing_key, no_wait, arguments),
            )
            return (yield from fut)

    def queue_purge(self, queue, *, no_wait=False,
                    method=spec.QueuePurge,
                    reply_method=spec.QueuePurgeOK):
        """Remove all messages from a queue which are not awaiting
        acknowledgment.

        :param queue: The name of the queue to purge.
        :keywork no_wait: If True, the client will not wait for the reply.
        """
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (0, queue, no_wait))
            return (yield from fut)

    def queue_delete(self, queue, *, if_unused=False, if_empty=False,
                     no_wait=False, method=spec.QueueDelete,
                     reply_method=spec.QueueDeleteOK):
        """Delete a queue. When a queue is deleted any pending messages
        are sent to a dead-letter queue if this is defined in the server
        configuration, and all consumers on the queue are cancelled.

        :param queue: The name of the queue to delete.
        :keyword if_unused: If True, the server will only delete the queue
                            if it has no consumers.
        :keyword if_empty: If True, the server will only delete the queue
                           if it has no messages.
        :keywork no_wait: If True, the client will not wait for the reply.
        """
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (0, queue, if_unused, if_empty, no_wait))
            return (yield from fut)

    # Work with Basic

    def basic_qos(self, prefetch_size, prefetch_count, *, a_global=False,
                  method=spec.BasicQos):
        """Request a specific quality of service.

        :param prefetch_size: Prefetch window in octets.
        :param prefetch_count: Prefetch window im messages.
        :keyword a_global: If False, the QoS will be applied only to new
                           consumers on the current channel.
        """
        self.send_method(method, (prefetch_size, prefetch_count, a_global))
        yield  # to make API more consistent

    def basic_consume(self, queue, *, consumer_tag=None, no_local=False,
                      no_ack=True, exclusive=False, no_wait=False,
                      arguments=None, consumer_callback=None, on_cancel=None,
                      method=spec.BasicConsume,
                      reply_method=spec.BasicConsumeOK):
        """Ask the server to start a "consumer", which is a transient request
        for messages from a specific queue.

        :param queue: The queue name to consume from.
        :keyword consumer_tag: The identified for the consumer. If is None,
                               the server will generate a unique tag.
        :keyword no_local: If True, the server will not send messages to the
                           connection that published them.
        :keyword no_ack: If True, the server does not expect acknowledgments
                         for messages.
        :keyword exclusive: Request exclusive consumer access,
                            meaning only this consumer can access the queue.
        :keyword no_wait: If True, the client will not wait for the reply.
        :keyword arguments: Additional arguments.
        :keyword consumer_callback: Optional callback that will be called
                                    when a message is received.
        :keyword on_cancel: Optional callback that will be called when
                            the consumer will be cancelled.
        """
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, queue, consumer_tag, no_local, no_ack, exclusive,
                 no_wait, arguments),
            )
            proposed_tag = yield from fut

        # In case consumer_tag was not provided
        if not no_wait and not consumer_tag:
            consumer_tag = proposed_tag

        assert consumer_tag

        self.consumer_callbacks[consumer_tag] = consumer_callback

        if on_cancel:
            self.cancel_callbacks[consumer_tag] = on_cancel
        if no_ack:
            self.no_ack_consumers.add(consumer_tag)
        return consumer_tag

    def basic_cancel(self, consumer_tag, *, no_wait=False,
                     method=spec.BasicCancel,
                     reply_method=spec.BasicCancelOK):
        """Cancel a consumer.

        :param consumer_tag: The identified for the consumer.
        :keyword no_wait: If True, the client will not wait for the reply.
        """
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (consumer_tag, no_wait))
            return (yield from fut)

    def _basic_publish(self, message, exchange, routing_key, *,
                       mandatory=False, immediate=False,
                       method=spec.BasicPublish):
        """Publish a message to a specific exchange.

        :param message: The message to publish.
        :param exchange: The name of the exchange to publish to.
        :param routing_key: The routing key for the message.
        :keyword mandatory: If True, the server will return an unroutable
                            message with a BasicReturn method.
                            Ife False, the server silently drops the message.
        :keyword immediate: If True, the server will return an undeliverable
                            message with a BasicReturn method. If False,
                            the server will queue the message, but with no
                            guarantee that it will ever be consumed.
        """
        self.send_method(
            method,
            (0, exchange, routing_key, mandatory, immediate),
            message=message,
            frame_size=self.connection.frame_max,
        )
        yield  # to make API more consistent

    # In case connection does not confirm publications
    basic_publish = _basic_publish

    def basic_publish_confirm(self, *args, reply_method=spec.BasicAck,
                              **kwargs):
        """Publish a message and wait for the confirm.
        See `basic_publish` for available arguments.
        """
        if not self._confirm_selected:
            self._confirm_selected = True
            self.confirm_select()
        with self.handler.wait_for(reply_method) as fut:
            self._basic_publish(*args, **kwargs)
            return (yield from fut)

    def basic_get(self, queue, *, no_ack=True,
                  method=spec.BasicGet,
                  reply_method=(spec.BasicGetOK, spec.BasicGetEmpty)):
        """This method provides a direct access to the messages in a queue
        using a synchronous dialogue.

        :param queue: The name of the queue to get a message from.
        :keyword no_ack: If True, the server does not expect acknowledgments
                         for messages.
        """
        with self.handler.wait_for(*reply_method) as fut:
            self.send_method(method, (0, queue, no_ack))
            res = yield from fut
            if not isinstance(res, BasicMessage):
                raise QueueEmpty('Queue "{}" is empty!')
            return res

    def basic_ack(self, delivery_tag, *, multiple=False,
                  method=spec.BasicAck):
        """Acknowledge one or more messages delivered via the BasicDeliver
        or BasicGetOK methods.

        :param delivery_tag: The server-assigned and channel-specific
                             delivery tag.
        :keyword multiple: If True, the `delivery_tag` is treated as
                           "up to and including". If False, the `delivery_tag`
                           refers to a single message. If True, and
                           `delivery_tag` is 0 (zero), this indicates
                           acknowledgment of all outstanding messages.
        """
        self.send_method(method, (delivery_tag, multiple))
        yield  # to make API more consistent

    def basic_reject(self, delivery_tag, *, requeue=False,
                     method=spec.BasicReject):
        """Reject a message.

        :param delivery_tag: The server-assigned and channel-specific
                             delivery tag.
        :keyword requeue: If False, the message will be redelivered
                          to the original recepient. If True, the server
                          will attempt to requeue the message.
        """
        self.send_method(method, (delivery_tag, requeue))
        yield  # to make API more consistent

    def basic_recover_async(self, *, requeue=False,
                            method=spec.BasicRecoverAsync):
        """Deprecated, use `basic_recover`"""
        self.send_method(method, (requeue,))
        yield  # to make API more consistent

    def basic_recover(self, *, requeue=False, method=spec.BasicRecover,
                      reply_method=spec.BasicRecoverOK):
        """Ask the server to redeliver all unacknowledged messages
        on a specific channel.

        :keyword requeue: If False, the message will be redelivered
                          to the original recepient. If True, the server
                          will attempt to requeue the message.
        """
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (requeue,))
            return (yield from fut)

    # Work with transactions

    def tx_select(self, *, method=spec.TxSelect, reply_method=spec.TxSelectOK):
        """Set the channel to use standard transactions."""
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, ())
            return (yield from fut)

    def tx_commit(self, *, method=spec.TxCommit, reply_method=spec.TxCommitOK):
        """Commit all message publications and acknowledgments performed
        in the current transactoin.
        """
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, ())
            return (yield from fut)

    def tx_rollback(self, *, method=spec.TxRollback,
                    reply_method=spec.TxRollbackOK):
        """Abandon all message publications and acknowledgments performed
        in the current transaction.
        """
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, ())
            return (yield from fut)

    # Work with confirms

    def confirm_select(self, *, nowait=False, method=spec.ConfirmSelect,
                       reply_method=spec.ConfirmSelectOK):
        """Set the channel to use publisher acknowledgments."""
        reply_method = None if nowait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (nowait,))
            return (yield from fut)
