# -*- coding: utf-8 -*-
"""Helper script to generate methods from xml spec."""
import os
import sys
from datetime import datetime
from xml.etree import ElementTree
from collections import OrderedDict


HEADER = '''\
# -*- coding: utf-8 -*-
"""
colibri.spec
~~~~~~~~~~~~

Implementation of AMQP specification.

This file was generated {} from {}.

"""
'''


IMPORTS = """\
from struct import pack, unpack

from .serialization import (dumps, loads, BIT, OCTET, SHORT, LONG,
                            LONGLONG, SHORTSTR, LONGSTR, TABLE)

"""


BASE_CLASSES = '''\
def select_method(method_type, unpack=unpack):
    code = unpack('!HH', method_type)
    return METHODS[code]


class MethodMeta(type):

    def __hash__(cls):
        return hash(cls.method_type)

    def __eq__(cls, other):
        if not isinstance(other, tuple):
            other = other.method_type
        return cls.method_type == other


class BaseMethod(object, metaclass=MethodMeta):
    """Base class for all AMQP methods."""

    @classmethod
    def unpack(cls, raw):
        return loads(''.join(type_ for _, type_ in cls.field_info), raw)[0]

    @classmethod
    def pack(cls, payload):
        payload = list(payload)
        fmt, values = [], []
        for name, type_ in cls.field_info:
            fmt.append(type_)
            values.append(payload.pop(0))
        return pack('!HH', *cls.method_type) + dumps(''.join(fmt), values)

'''


def load_spec(filename):
    tree = ElementTree.parse(filename)
    constants = get_constants(tree)
    methods, code_to_method = generate_methods(tree)
    return constants, methods, code_to_method


def get_constants(tree):
    constants = []
    for elem in tree.findall('constant'):
        name = elem.attrib['name'].replace('-', '_').upper()
        value = int(elem.attrib['value'])
        constants.append('{} = {}'.format(name, value))
    return '# AMQP constants\n' + '\n'.join(constants)


def generate_methods(tree):

    classes = get_classes(tree)
    methods = []
    code_to_method = ['METHODS = {']
    for cname, (cid, cmethods) in classes.items():
        for mname, (mid, mfields, msupport, synchronous, mdoc) in cmethods.items():  # noqa
            name = cname + mname
            method_type = (cid, mid)

            mfields_items = list(mfields.items())
            if mfields_items:
                mfields = '(\n        {}\n    )'.format(
                    '\n        '.join("('{}', {}),".format(k, v)
                                      for k, v in mfields_items)
                )
            else:
                mfields = '()'
            methods.append('''\
class {name}(BaseMethod):
    """{mdoc}
    """
    method_type = {method_type}
    field_info = {mfields}
    synchronous = {synchronous}\n
'''.format(**{'name': name, 'method_type': method_type, 'mfields': mfields,
              'synchronous': synchronous, 'mdoc': mdoc}))
            code_to_method.append('    {!r}: {},'.format(method_type, name))
    code_to_method.append('}\n')
    return methods, code_to_method


def get_classes(tree):
    domain_types = {e.attrib['name']: e.attrib['type']
                    for e in tree.findall('domain')}
    classes = OrderedDict()
    for class_elem in tree.findall('class'):
        class_methods = OrderedDict()
        for method_elem in class_elem.findall('method'):

            fields = OrderedDict()
            for field_elem in method_elem.findall('field'):
                field_name = field_elem.attrib['name'].replace('-', '_')
                try:
                    field_type = field_elem.attrib['type']
                except KeyError:
                    field_type = domain_types[field_elem.attrib['domain']]
                fields[field_name] = field_type.upper()

            method_support = OrderedDict()
            for elem in method_elem.findall('chassis'):
                method_support[elem.attrib['name']] = elem.attrib['implement']

            doc = build_docstring(method_elem, fields)
            synchronous = 'synchronous' in method_elem.attrib

            method_id = int(method_elem.attrib['index'])
            method_name = (method_elem.attrib['name']
                           .capitalize()
                           .replace('-ok', 'OK')
                           .replace('-empty', 'Empty')
                           .replace('-async', 'Async'))
            class_methods[method_name] = (method_id, fields, method_support,
                                          synchronous, doc)

        class_id = int(class_elem.attrib['index'])
        classes[class_elem.attrib['name'].capitalize()] = (class_id,
                                                           class_methods)
    return classes


def build_docstring(method_elem, fields):
    text = method_elem.find('doc').text.split()
    doc = []
    line = []
    first = True
    for word in text:
        if len(' '.join(line)) + len(word) > (70 if first else 74):
            first = False
            doc.append(' '.join(line))
            line = []
        line.append(word)
    if line:
        doc.append(' '.join(line))
    description = '\n    '.join(doc)
    args_from_fields = ['    {}: {}'.format(n, t) for n, t in fields.items()]
    if args_from_fields:
        args = '\n\n    Arguments:\n    {}'.format(
            '\n    '.join(args_from_fields)
        )
    else:
        args = ''
    return ''.join((description, args))


if __name__ == '__main__':
    proj_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    len_argv = len(sys.argv)
    if len_argv > 1:
        spec_source = sys.argv[1]
    else:
        spec_source = proj_dir + '/spec/amqp0-9-1.extended.xml'
    spec_file = proj_dir + '/colibri/spec.py'
    CONSTANTS, METHODS, CODE_TO_METHOD = load_spec(spec_source)
    with open(spec_file, 'w') as f:
        f.write('\n'.join((HEADER.format(datetime.now(),
                                         os.path.basename(spec_source)),
                           IMPORTS,
                           BASE_CLASSES,
                           '\n'.join(METHODS),
                           '\n'.join(CODE_TO_METHOD),
                           CONSTANTS)))
